/*
 * autopilot.c:
 *
 * Copyright (c) 2001 James McKenzie <james@fishsoup.dhs.org>,
 * All rights reserved.
 * based on code of same name 	
 * Copyright (C) 2000 Leigh Morresi (leigh@signal-x.com)
 */


/*
 * Monifications - look at CTS line as this is asserted by the
 * serial driver throughout the duration (DSR if often 
 * disconnected to stop cradle drain). If CTS is deasserted 
 * it signifies either handshaking - or that the hot sync has  
 * failed. We terminate the child and go back to waiting if
 * timeout reads spaced delay apart return CTS deasserted 
 */

static char rcsid[] =
  "$Id: autopilot.c,v 1.3 2001/09/14 12:11:50 root Exp root $";


#define SHELL_NAME "sh"
#define SHELL_PATH "/bin/sh"

int bauds[] = {
  9600,
  57600,
  115200,
};

#define  NBAUDS (sizeof(bauds)/sizeof(int))



/*
 * $Log: autopilot.c,v $
 * Revision 1.3  2001/09/14 12:11:50  root
 * *** empty log message ***
 *
 * Revision 1.2  2001/09/14 10:26:01  root
 * *** empty log message ***
 *
 * Revision 1.1  2001/09/14 10:25:54  root
 * Initial revision
 *
 */

/*
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 */

#ifndef PALMDEVICE
#define PALMDEVICE "/dev/pilot"
#endif

#include <syslog.h>
#include <stdlib.h>
#include <stdio.h>
#include <unistd.h>
#include <sys/ioctl.h>
#include <fcntl.h>
#include <signal.h>
#include <errno.h>
#include <string.h>
#include <sys/wait.h>
#include <termios.h>
#include <sys/types.h>
#include <sys/time.h>


/* global */
char *device = PALMDEVICE;
char *ncommand = NULL;
char *hcommand = NULL;
int delay = 2;
int timeout = 10;

/* print command usage and exit */
void
usage ()
{
  fprintf (stderr, "usage: autopilot [-p port] -n <pppcmd> -h <synccmd>\n");
  fprintf (stderr,
           "  -m pppd command to run if we see ppp packets [baud rate appended]\n");
  fprintf (stderr, "  -h hotsync command to run if we see pisock packets\n");
  fprintf (stderr, "  -p port to listen on." " (default %s)\n", PALMDEVICE);
  fprintf (stderr, "  -n interval between scan for pilot in seconds."
           " (default %d)\n", delay);
  fprintf (stderr, "  -t number of intervals to detect timeout."
           " (default %d)\n", timeout);
  exit (1);
}

/* handle command line options */
void
parse_args (int argc, char **argv)
{
  const char *flags = "m:t:n:h:p:";
  int c;
  extern char *optarg;

  while ((c = getopt (argc, argv, flags)) != EOF)
    {
      switch (c)
        {
        case 'n':
          delay = atoi (optarg);
          break;
        case 't':
          timeout = atoi (optarg);
          break;
        case 'm':
          ncommand = optarg;
          break;
        case 'h':
          hcommand = optarg;
          break;
        case 'p':
          device = optarg;
          break;
        case '?':
          usage ();
          exit (1);
          break;
        }
    }

}


  int pid=0;

static void medea(int i)
{
if (pid) {
	kill(pid,SIGKILL);
	pid=0;
}
  signal(SIGTERM,SIG_DFL);
  exit(1);
}


void
run_child (int fd,char *command)
{
  unsigned int arg;
  int deasserted = 0;
  int status;

  syslog (LOG_ERR, "running %s on %s", command,device);
  switch (pid = fork ())
    {
    case 0:
      {
        char *args[4];

        close (fd);

        args[0] = SHELL_NAME;
        args[1] = "-c";
        args[2] = command;
        args[3] = NULL;

        signal (SIGINT, SIG_DFL);
        signal (SIGQUIT, SIG_DFL);

        execv (SHELL_PATH, args);
        syslog (LOG_ERR, "run_child: execve %s -c %s failed %m", SHELL_PATH,
                command);
        exit (-1);
      }

      break;
    case -1:
      syslog (LOG_ERR, "run_child: fork failed %m");
      return;
    }


  syslog (LOG_ERR, "run_child: launched %d", pid);
  for (;;)
    {

      switch (waitpid (pid, &status, WNOHANG))
        {
        case 0:
          break;
        case -1:
          /* Possible race condition gets us here */
          syslog (LOG_ERR, "run_child: waitpid %m");
	  pid=0;
          return;
        default:
          syslog (LOG_ERR, "run_child: collected %d", pid);
	  pid=0;
          return;
        }




      if (ioctl (fd, TIOCMGET, &arg))
        syslog (LOG_ERR, "TIOCMGET failed %m");

      if (arg & TIOCM_CTS)
        {
          deasserted = 0;
        }
      else
        {
          deasserted++;
        }


      if (deasserted == timeout)
        {
          /*Kill child, waitpid will clean up and quit this loop */
          syslog (LOG_ERR, "Zapping child %d", pid);
          kill (pid,SIGTERM);
        }

      if (deasserted == (2 * timeout))
        {
          /*Kill child, waitpid will clean up and quit this loop */
          syslog (LOG_ERR, "Zapping child %d", pid);
          kill (pid,SIGKILL);
        }

/*
      if (deasserted == (3 * timeout))
        {
        char *args[4];
        syslog (LOG_ERR, "Self destruct initiated...", pid);
        args[0] = SHELL_NAME;
        args[1] = "-c";
        args[2] = "/software/scripts/sr";
        args[3] = NULL;

        signal (SIGINT, SIG_DFL);
        signal (SIGQUIT, SIG_DFL);

        execv (SHELL_PATH, args);
        syslog (LOG_ERR, "run_child: execve %s -c %s failed %m", SHELL_PATH,
                command);
        exit (-1);
        }
*/

      sleep (delay);

    }
}



int
inttospd (int s)
{
  switch (s)
    {
    case 50:
      return B50;
    case 75:
      return B75;
    case 110:
      return B110;
    case 134:
      return B134;
    case 150:
      return B150;
    case 200:
      return B200;
    case 300:
      return B300;
    case 600:
      return B600;
    case 1200:
      return B1200;
    case 1800:
      return B1800;
    case 2400:
      return B2400;
    case 4800:
      return B4800;
    case 9600:
      return B9600;
    case 19200:
      return B19200;
    case 38400:
      return B38400;
    case 57600:
      return B57600;
    case 115200:
      return B115200;
    case 230400:
      return B230400;
    case 460800:
      return B460800;
    case 500000:
      return B500000;
    case 576000:
      return B576000;
    case 921600:
      return B921600;
    case 1000000:
      return B1000000;
    case 1152000:
      return B1152000;
    case 1500000:
      return B1500000;
    case 2000000:
      return B2000000;
    case 2500000:
      return B2500000;
    case 3000000:
      return B3000000;
    case 3500000:
      return B3500000;
    case 4000000:
      return B4000000;
    default:
      return 0;
    }
}

int
spdtoint (int s)
{
  switch (s)
    {
    case B50:
      return 50;
    case B75:
      return 75;
    case B110:
      return 110;
    case B134:
      return 134;
    case B150:
      return 150;
    case B200:
      return 200;
    case B300:
      return 300;
    case B600:
      return 600;
    case B1200:
      return 1200;
    case B1800:
      return 1800;
    case B2400:
      return 2400;
    case B4800:
      return 4800;
    case B9600:
      return 9600;
    case B19200:
      return 19200;
    case B38400:
      return 38400;
    case B57600:
      return 57600;
    case B115200:
      return 115200;
    case B230400:
      return 230400;
    case B460800:
      return 460800;
    case B500000:
      return 500000;
    case B576000:
      return 576000;
    case B921600:
      return 921600;
    case B1000000:
      return 1000000;
    case B1152000:
      return 1152000;
    case B1500000:
      return 1500000;
    case B2000000:
      return 2000000;
    case B2500000:
      return 2500000;
    case B3000000:
      return 3000000;
    case B3500000:
      return 3500000;
    case B4000000:
      return 4000000;
    default:
      return 0;
    }
}


void
do_fd (int fd)
{
  struct termios tcn;
  int i;

  tcgetattr (fd, &tcn);

  cfmakeraw (&tcn);

  tcn.c_oflag = 0;
  tcn.c_iflag = IGNBRK | IGNPAR;

  tcn.c_cflag = CREAD | CLOCAL | CS8;

  tcn.c_lflag = 0;

  (void) cfsetspeed (&tcn, B9600);

  for (i = 0; i < 16; i++)
    tcn.c_cc[i] = 0;

  tcsetattr (fd, TCSANOW, &tcn);
}

void
set_speed (int fd, int speed)
{
  int psd = inttospd (speed);
  struct termios tcn;

  tcgetattr (fd, &tcn);
  (void) cfsetspeed (&tcn, psd);
  tcsetattr (fd, TCSAFLUSH, &tcn);
}

int open_fd(char *port)
{
  int fd = open (port, O_RDONLY | O_NDELAY | O_NOCTTY);
  long i;

  if (fd < 0)
    {
      fprintf (stderr, "Can't open device %s: %s\n", device,
               strerror (errno));
      usage ();
      exit (1);
    }

  do_fd (fd);

  if ((i = fcntl (fd, F_GETFL, 0)) != -1)
    {
      i &= ~O_NONBLOCK;
      fcntl (fd, F_SETFL, i);
    }

   return fd;
}


int
get_speed (int fd)
{
  int spd;
  struct termios tcn;

  tcgetattr (fd, &tcn);
  spd = cfgetispeed (&tcn);
  return spdtoint (spd);
}

int
main (int argc, char *argv[])
{
  int fd;
  int i;

  if (getenv ("PILOTPORT") != NULL)
    device = getenv ("PILOTPORT");

  parse_args (argc, argv);

  if (!hcommand)
    {
      fprintf (stderr,
               "You must specify a hcommand to run with the -h option.\n");
      usage ();
      exit (1);
    }


  if (!ncommand)
    {
      fprintf (stderr,
               "You must specify a ncommand to run with the -n option.\n");
      usage ();
      exit (1);
    }

  if (!device)
    {
      fprintf (stderr, "No device set PILOTPORT or use -p option");
      usage ();
      exit (1);
    }

  fd=open_fd(device);

#ifndef DEBUG
  switch (fork ())
    {
    case 0:
      break;
    case -1:
      perror ("fork failed");
      exit (1);
    default:
      exit (0);
    }

  daemon (0, 0);
#endif

  openlog ("ap-ident", LOG_CONS, LOG_DAEMON);

  signal (SIGCHLD, SIG_IGN);
  signal(SIGTERM,medea);

  sleep(1);

  while (1)
    {

      for (i = 0; i < NBAUDS; ++i)
        {
          struct timeval then, now;
          fd_set readfds;
          long d;
              unsigned char c = 0, oc;

          set_speed (fd, bauds[i]);
          gettimeofday (&then, NULL);
          then.tv_sec += 2;

          while (1)
            {
		if (fd<0) {
			fd=open_fd(device);
		}

              gettimeofday (&now, NULL);

              d = then.tv_sec - now.tv_sec;
              d *= 1000000;
              d += then.tv_usec - now.tv_usec;

              if (d < 0)
                break;
              now.tv_sec = d / 1000000;
              now.tv_usec = d - (now.tv_sec * 1000000);


              FD_ZERO (&readfds);
              FD_SET (fd, &readfds);


              if (select (fd + 1, &readfds, NULL, NULL, &now))
                {
	  	  oc = c;
                  read (fd, &c, 1);

                  if ((oc == 0x7e) && (c == 0xff))
                    {
		      char command[1024],baud[80];
			syslog(LOG_ERR,"ppp connection at %d baud",get_speed(fd));
			strcpy(command,ncommand);
			sprintf(baud," %d",get_speed(fd));
			strcat(command,baud);
			run_child(fd,command);
			close(fd);
			fd=-1;
			break;
			
                    }
                  if ((oc == 0xbe) && (c == 0xef))
                    {
			syslog(LOG_ERR,"pi-sock connection at %d baud",get_speed(fd));
			run_child(fd,hcommand);
			close(fd);
			fd=-1;
			break;
                    }

                }

            }
        }
    }

  return 0;
}
